#include "Buffer.h"
#include "Math.h"
#include "Debug.h"
#include <stdexcept>
#include <math.h>
#include <new>

//////////////////
// Constructors //
//////////////////

buffer::buffer() : data(nullptr), sizeUsed(0), sizeAllotted(0)
{
	bufTitle[4] = NULL;
}

buffer::buffer(const char* bufferTitle) : data(nullptr), sizeUsed(0), sizeAllotted(0)
{
	strcpy_s(bufTitle, 5, bufferTitle);
	bufTitle[4] = NULL;
}

////////////////
// Destructor //
////////////////

buffer::~buffer()
{
	if ( data )
		delete[] data;
}

///////////////
// Accessors //
///////////////

ULONG buffer::size()
{
	if ( this )
		return sizeUsed;
	else
		return 0;
}

char* buffer::title()
{
	if ( this )
		return bufTitle;
	else
		return 0;
}

template <typename valueType>
bool buffer::get(valueType &dest, ULONG location)
{
	if ( this )
	{
		if ( location+sizeof(valueType) <= sizeUsed )
		{
			dest = (valueType &)data[location];
			return true;
		}
	}
	return false;
}
template bool buffer::get<BYTE>(BYTE &dest, ULONG location);
template bool buffer::get<USHORT>(USHORT &dest, ULONG location);
template bool buffer::get<ULONG>(ULONG &dest, ULONG location);
template bool buffer::get<LONG>(LONG &dest, ULONG location);

bool buffer::getBit(bool &dest, ULONG location, ULONG bitNum)
{
	if ( this )
	{
		location += bitNum / 8;

		if ( location < sizeUsed )
		{
			dest = ( (char(1) << (bitNum%8)) & data[location] ) != 0;
			// AND the bit given by bitNum
			return true;
		}
	}
	return false;
}

bool buffer::getString(char* dest, ULONG location, ULONG size)
{
	if ( this )
	{
		if ( location+size < sizeUsed )
		{
			for ( ULONG i=0; i<size; i++ )
				dest[i] = data[i+location];

			return true;
		}
	}
	return false;
}

/**
bool buffer::getPtr(char* &dest, ULONG location)
{
	if ( this )
	{
		if ( location < sizeUsed )
		{
			dest = &data[location];
			return true;
		}
	}
	return false;
}
*/

void* buffer::getPtr(ULONG location)
{
	if ( this && location < sizeUsed )
		return (void*)&data[location];
	else
		return nullptr;
}

void* buffer::getPtr(ULONG location, ULONG sizeRequested)
{
	if ( this && location+sizeRequested <= sizeUsed )
		return (void*)&data[location];
	else
		return nullptr;
}

template <typename valueType>
bool buffer::getPtr(valueType* &dest, ULONG location, ULONG sizeRequested)
{
	if ( this )
	{
		if ( location+sizeRequested <= sizeUsed )
		{
			dest = (valueType*)&data[location];
			return true;
		}
	}
	return false;
}
template bool buffer::getPtr<char>(char* &dest, ULONG location, ULONG sizeRequested);
template bool buffer::getPtr<BYTE>(BYTE* &dest, ULONG location, ULONG sizeRequested);
template bool buffer::getPtr<ULONG>(ULONG* &dest, ULONG location, ULONG sizeRequested);

template <typename valueType>
valueType buffer::get(ULONG location)
{
	if ( this )
	{
		if ( location+sizeof(valueType) <= sizeUsed )
			return (valueType &)data[location];
	}
	return 0;
}
template BYTE buffer::get<BYTE>(ULONG location);
template USHORT buffer::get<USHORT>(ULONG location);
template ULONG buffer::get<ULONG>(ULONG location);

bool buffer::getBit(ULONG location, ULONG bitNum)
{
	if ( this )
	{
		location += bitNum / 8;

		if ( location < sizeUsed )
		{
			return ( (char(1) << (bitNum%8)) & data[location] ) != 0;
			// AND the bit given by bitNum
		}
	}
	return false;
}

///////////////
// Searching //
///////////////

bool buffer::has(char character, ULONG location)
{
	if ( this )
	{
		if ( location < sizeUsed )
		{
			if ( data[location] == character )
				return true;
		}
	}
	return false;
}

bool buffer::has(const char* str, ULONG location, ULONG size)
{
	if ( this )
	{
		if ( location + size <= sizeUsed )
		{
			for ( ULONG i=0; i<size; i++ )
			{
				if ( data[i+location] != str[i] )
					return false;
			}
			return true;
		}
	}
	return false;
}

bool buffer::getNext(char character, ULONG start, ULONG &dest)
{
	if ( this )
	{
		if ( start < sizeUsed )
		{
			for ( ULONG i=start; i<sizeUsed; i++ )
			{
				if ( data[i] == character )
				{
					dest = i;
					return true;
				}
			}
		}
	}
	return false;
}

//////////////
// Mutators //
//////////////

template <typename valueType>
bool buffer::add(valueType value)
{
	if ( this )
	{
		if ( sizeUsed+sizeof(valueType) > sizeAllotted )
		{
			try {
				resize(sizeof(valueType), true);
			} catch ( const BadResize &e ) {
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to increase size of %s buffer: %s", bufTitle, e.what());
				strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.add(value)");
				return false;
			}
		}

		(valueType &)data[sizeUsed] = value;
		sizeUsed += sizeof(valueType);

		return true;
	}
	return false;
}
template bool buffer::add<BYTE>(BYTE value);
template bool buffer::add<USHORT>(USHORT value);
template bool buffer::add<ULONG>(ULONG value);

template <typename valueType>
bool buffer::add(valueType value, ULONG amount)
{
	if ( this )
	{
		if ( sizeUsed+amount*sizeof(valueType) >= sizeAllotted )
		{
			try {
				resize(amount*sizeof(valueType), true);
			} catch ( const BadResize &e ) {
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to increase size of %s buffer: %s", bufTitle, e.what());
				strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.add(value, amount)");
				return false;
			}
		}

		for ( ULONG i=sizeUsed+(amount*sizeof(valueType))-sizeof(valueType); i>=sizeUsed; i-=sizeof(valueType) )
			(valueType &)data[i] = value;

		sizeUsed += amount*sizeof(valueType);

		return true;
	}
	return false;
}
template bool buffer::add<BYTE>(BYTE value, ULONG amount);
template bool buffer::add<USHORT>(USHORT value, ULONG amount);

bool buffer::addStr(const char* chunk, ULONG chunkSize)
{
	if ( this )
	{
		if ( sizeUsed+chunkSize >= sizeAllotted )
		{
			try {
				resize(chunkSize, true);
			} catch ( const BadResize &e ) {
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to increase size of %s buffer: %s", bufTitle, e.what());
				strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.add(chunk, chunkSize)");
				return false;
			}
		}

		for ( ULONG i=sizeUsed+chunkSize; i>sizeUsed; i-- )
			data[i-1] = chunk[i-sizeUsed-1];

		sizeUsed += chunkSize;

		return true;
	}
	return false;
}

template <typename valueType>
bool buffer::insert(ULONG location, valueType value)
{
	if ( this )
	{
		if ( location+sizeof(valueType) < sizeUsed )
		{
			if ( sizeUsed+sizeof(valueType) > sizeAllotted )
			{
				try {
					resize(sizeof(valueType), true);
				} catch ( const BadResize &e ) {
					sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to increase size of %s buffer: %s", bufTitle, e.what());
					strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.insert(location, character)");
					return false;
				}
			}

			for ( ULONG i = sizeUsed+sizeof(valueType)-1; i > location; i-- )
				data[i] = data[i-1];
		
			(valueType &)data[location] = value;
			sizeUsed++;
			return true;
		}
	}
	return false;
}
template bool buffer::insert<char>(ULONG location, char value);

bool buffer::insertStr(ULONG startLocation, const char* chunk, ULONG chunkSize)
{
	if ( this )
	{
		if ( startLocation < sizeUsed )
		{
			if ( sizeUsed+chunkSize > sizeAllotted )
			{
				try {
					resize(chunkSize, true);
				} catch ( const BadResize &e ) {
					sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to increase size of %s buffer: %s", bufTitle, e.what());
					strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.insert(startLocation, chunk, chunkSize)");
					return false;
				}
			}

			for ( ULONG i=sizeUsed+chunkSize; i>startLocation+chunkSize; i-- )
				data[i-1] = data[i-chunkSize-1];

			for ( ULONG i=startLocation+chunkSize-1; i>startLocation; i-- )
				data[i-1] = chunk[i-startLocation-1];

			sizeUsed += chunkSize;
			return true;
		}
	}
	return false;
}

template <typename valueType>
bool buffer::replace(ULONG location, valueType value)
{
	if ( this )
	{
		if ( location+sizeof(valueType)-1 < sizeUsed )
		{
			(valueType &)data[location] = value;
			return true;
		}
	}
	return false;
}
template bool buffer::replace<char>(ULONG location, char value);
template bool buffer::replace<BYTE>(ULONG location, BYTE value);
template bool buffer::replace<USHORT>(ULONG location, USHORT value);

bool buffer::replaceStr(ULONG startLocation, const char* chunk, ULONG chunkSize)
{
	if ( this )
	{
		if ( startLocation+chunkSize < sizeUsed )
		{
			for ( ULONG i=0; i<chunkSize; i++ )
				data[i+startLocation] = chunk[i];

			return true;
		}
	}
	return false;
}

bool buffer::replaceStr(ULONG startLocation, ULONG initialSize, const char* chunk, ULONG chunkSize)
{
	if ( this )
	{
		if ( startLocation+initialSize-1 < sizeUsed )
		{
			signed long long difference = chunkSize-initialSize;

			for ( ULONG i=0; i<chunkSize; i++ )
				data[i+startLocation] = chunk[i];

			if ( difference > 0 )
			{
				try {
					resize(difference, true);
				} catch ( const BadResize &e ) {
					sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to increase size of %s buffer: %s", bufTitle, e.what());
					strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.replace(startLocation, initialSize, chunk, chunkSize)");
					return false;
				}

				for ( ULONG i=ULONG(sizeUsed+difference); i>startLocation+chunkSize; i-- )
					data[i-1] = data[i-difference-1];

				for ( ULONG i=chunkSize; i>initialSize; i-- )
					data[i+startLocation-1] = chunk[i-1];
			}
			else if ( difference < 0 )
			{
				for ( ULONG i=startLocation+initialSize; i<sizeUsed; i++ )
					data[i+difference] = data[i];

				try {
					resize(difference, false);
				} catch ( const BadResize &e ) {
					sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to decrease size of %s buffer: %s", bufTitle, e.what());
					strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.replace(startLocation, initialSize, chunk, chunkSize)");
					return false;
				}
			}

			sizeUsed += (ULONG)difference;
			return true;
		}
	}
	return false;
}

bool buffer::overwrite(const char* chunk, ULONG chunkSize)
{
	flush();
	if ( !resize(chunkSize, true) )
		return false;

	for ( ULONG i=0; i<chunkSize; i++ )
		data[i] = chunk[i];

	sizeUsed = chunkSize;

	return true;
}

template <typename valueType>
bool buffer::del(ULONG location)
{
	if ( this )
	{
		if ( location+sizeof(valueType) <= sizeUsed )
		{
			for ( ULONG i=location+sizeof(valueType); i<sizeUsed; i++ )
				data[i-sizeof(valueType)] = data[i];

			try {
				resize(-(int)sizeof(valueType), false);
			} catch ( const BadResize &e ) {
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to decrease size of %s buffer: %s", bufTitle, e.what());
				strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.del(location)");
				return false;
			}
			return true;
		}
	}
	return false;
}
template bool buffer::del<BYTE>(ULONG location);

bool buffer::del(ULONG startLocation, ULONG endLocation)
{
	if ( this )
	{
		if ( endLocation < sizeUsed && startLocation < endLocation )
		{
			signed long long difference = endLocation-startLocation;
			for ( ULONG i=endLocation; i<sizeUsed; i++ )
				data[i-difference] = data[i];

			try {
				resize(-difference, false);
			} catch ( const BadResize &e ) {
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to decrease size of %s buffer: %s", bufTitle, e.what());
				strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.del(startLocation, endLocation)");
				return false;
			}
			return true;
		}
	}
	return false;
}

////////////////
// IO Methods //
////////////////

void buffer::write(FILE* pFile)
{
	if ( this )
	{
		fwrite(bufTitle, 4, 1, pFile);
		fwrite(&sizeUsed, 4, 1, pFile);
		if ( sizeUsed > 0 )
			fwrite(data, sizeUsed, 1, pFile);
	}
}

bool buffer::extract(buffer &buf, ULONG &position)
{
	if ( this )
	{
		if ( sizeUsed > 0 ) // Specified buffer already contains data!
			flush(); // Clear the buffer: StarCraft only reads the last instance of the given section

		//if ( !buf.get<ULONG>((ULONG &)bufTitle, position) )
		//	return false;

		position += 4;

		ULONG bufSize;
		if ( !buf.get<ULONG>(bufSize, position) )
			return false;

		position += 4;

		if ( position+bufSize > buf.sizeUsed )
			return false;

		try {
			resize(bufSize, false);
		} catch ( const BadResize &e ) {
			sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to increase size of %s buffer: %s", bufTitle, e.what());
			strcpy_s(LastErrorLoc, MAX_ERROR_LENGTH, "buffer.extract(buf, position)");
			return false;
		}

		sizeUsed = 0;
		sizeAllotted = bufSize;

		addStr(&buf.data[position], bufSize);

		position += bufSize;

		return true;
	}
	return false;
}

//////////
// Misc //
//////////

bool buffer::setSize(ULONG size)
{
	if ( this )
	{
		if ( size >= sizeUsed )
		{
			resize(size-sizeAllotted, false);
			return true;
		}
	}
	return false;
}

void buffer::flush()
{
	if ( this )
	{
		if ( data )
		{
			delete[] data;
			data = nullptr;
		}

		sizeUsed = 0;
		sizeAllotted = 0;
	}
}

//////////////////
// Priv Methods //
//////////////////

bool buffer::resize(signed long long sizeChange, bool multiplySize)
{
	if ( this )
	{
		if ( sizeChange > 0 && multiplySize )
			sizeAllotted = (ULONG)(sizeAllotted*1.2+sizeChange); // Increase allotted size by 20% + required space
		else
			sizeAllotted = (ULONG)(sizeAllotted+sizeChange); // Decrease allotted size by amount specified

		if ( sizeUsed > 0 )
		{
			char* newBuffer;
			try {
				newBuffer = new char[sizeAllotted];
			} catch ( std::bad_alloc ) {
				throw(const BadResize("could not create a temporary buffer - you may have ran out of memory."));
			}

			if ( sizeUsed <= sizeAllotted )
				memcpy( newBuffer, data, sizeUsed*sizeof(char) );
			else
				memcpy( newBuffer, data, sizeAllotted*sizeof(char) );

			if ( data )
			{
				delete[] data;
				data = nullptr;
			}

			data = newBuffer;
		}
		else
		{
			if ( data )
			{
				delete[] data;
				data = nullptr;
			}

			if ( sizeAllotted > 0 )
			{
				try {
					data = new char[sizeAllotted];
				} catch ( std::bad_alloc ) {
					sizeAllotted = 0;
					throw(const BadResize("could not initialize buffer - you may have ran out of memory."));
				}
			}
		}
		return true;
	}
	return false;
}